@@ -11,7 +11,9 @@
 #endif
 #if __GLASGOW_HASKELL__ >= 702
 {-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE EmptyDataDecls #-}
 {-# LANGUAGE Trustworthy #-}
+{-# LANGUAGE TypeFamilies #-}
 #endif
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 ----------------------------------------------------------------------------
@@ -33,7 +35,11 @@ module Data.Proxy
     , KProxy(..)
     ) where
 
-import Control.Applicative (Applicative(..))
+import Control.Applicative (Applicative(..), Alternative(..))
+import Control.Monad (MonadPlus(..))
+#if MIN_VERSION_base(4,4,0)
+import Control.Monad.Zip (MonadZip(..))
+#endif
 #ifdef MIN_VERSION_deepseq
 import Control.DeepSeq (NFData(..))
 #endif
@@ -45,7 +51,7 @@ import Data.Monoid
 import GHC.Arr (unsafeIndex, unsafeRangeSize)
 import Data.Data
 #if __GLASGOW_HASKELL__ >= 702
-import GHC.Generics (Generic)
+import GHC.Generics hiding (Fixity(..))
 #endif
 #endif
 
@@ -55,6 +61,27 @@ deriving instance Typeable Proxy
 data Proxy s = Proxy
 #if __GLASGOW_HASKELL__ >= 702
   deriving Generic
+
+-- We have to implement the Generic1 instance manually due to an old
+-- bug in GHC 7.6. This is mostly copied from the output of
+--
+-- deriving instance Generic1 Proxy
+--
+-- Compiled with -ddump-deriv on a more recent GHC.
+instance Generic1 Proxy where
+  type Rep1 Proxy = D1 ProxyMetaData (C1 ProxyMetaCons U1)
+  from1 Proxy = M1 (M1 U1)
+  to1 (M1 (M1 U1)) = Proxy
+
+data ProxyMetaData
+data ProxyMetaCons
+
+instance Datatype ProxyMetaData where
+  datatypeName _ = "Proxy"
+  moduleName   _ = "Data.Proxy"
+
+instance Constructor ProxyMetaCons where
+  conName _ = "Proxy"
 #endif
 #endif
 
@@ -142,6 +169,12 @@ instance Applicative Proxy where
     _ <*> _ = Proxy
     {-# INLINE (<*>) #-}
 
+instance Alternative Proxy where
+    empty = Proxy
+    {-# INLINE empty #-}
+    _ <|> _ = Proxy
+    {-# INLINE (<|>) #-}
+
 instance Monoid (Proxy s) where
     mempty = Proxy
     {-# INLINE mempty #-}
@@ -156,6 +189,18 @@ instance Monad Proxy where
     _ >>= _ = Proxy
     {-# INLINE (>>=) #-}
 
+instance MonadPlus Proxy where
+    mzero = Proxy
+    {-# INLINE mzero #-}
+    mplus _ _ = Proxy
+    {-# INLINE mplus #-}
+
+#if MIN_VERSION_base(4,4,0)
+instance MonadZip Proxy where
+    mzipWith _ _ _ = Proxy
+    {-# INLINE mzipWith #-}
+#endif
+
 instance Foldable Proxy where
     foldMap _ _ = mempty
     {-# INLINE foldMap #-}
